import csv

from django.contrib.admin.views.decorators import staff_member_required
from django.contrib.auth.decorators import login_required
from django.http import StreamingHttpResponse
from django.shortcuts import redirect, render

from allianceauth import NAME as site_header
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import TASKMONITOR_DATA_MAX_AGE
from .core import cached_reports
from .helpers import Echo
from .models import TaskLog

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


@login_required
@staff_member_required
def admin_taskmonitor_download_csv(request) -> StreamingHttpResponse:
    """Return all tasklogs as CSV file for download."""
    queryset = TaskLog.objects.order_by("pk")
    model = queryset.model
    exclude_fields = ("traceback",)

    logger.info("Preparing to export the task log with %s entries.", queryset.count())

    fields = [
        field
        for field in model._meta.fields + model._meta.many_to_many
        if field.name not in exclude_fields
    ]
    pseudo_buffer = Echo()
    writer = csv.writer(pseudo_buffer, delimiter=";")
    return StreamingHttpResponse(
        (writer.writerow(row) for row in queryset.csv_line_generator(fields)),
        content_type="text/csv",
        headers={"Content-Disposition": 'attachment; filename="tasklogs.csv"'},
    )


@login_required
@staff_member_required
def admin_taskmonitor_reports(request):
    """Show the reports page."""
    context = {
        "title": "Reports",
        "site_header": site_header,
        "cl": {"opts": TaskLog._meta},
        "data_max_age": TASKMONITOR_DATA_MAX_AGE,
    }
    context.update(cached_reports.data())
    return render(request, "admin/taskmonitor/tasklog/reports.html", context)


@login_required
@staff_member_required
def admin_taskmonitor_reports_clear_cache(request):
    """Reload the reports page with cleared cache."""
    cached_reports.clear_cache()
    return redirect("taskmonitor:admin_taskmonitor_reports")



from app_utils.django import clean_setting

TASKMONITOR_DATA_MAX_AGE = clean_setting("TASKMONITOR_DATA_MAX_AGE", 24)
"""Max age of logged tasks in hours. Older logs be deleted automatically."""

TASKMONITOR_HOUSEKEEPING_FREQUENCY = clean_setting(
    "TASKMONITOR_HOUSEKEEPING_FREQUENCY", 15
)
"""Frequency of house keeping runs in minutes."""

TASKMONITOR_REPORTS_MAX_AGE = clean_setting("TASKMONITOR_REPORTS_MAX_AGE", 15)
"""Max age of cached reports in minutes."""

TASKMONITOR_REPORTS_MAX_TOP = clean_setting("TASKMONITOR_REPORTS_MAX_TOP", 20)
"""Max items to show in the top reports. e.g. 10 will shop the top ten items."""



from unittest.mock import patch

from django.test import TestCase

from taskmonitor.models import TaskLog

from .factories import TaskLogFactory

MODELS_PATH = "taskmonitor.models"


class TestManagerCreateFromTask(TestCase):
    def test_should_create_from_failed_task(self):
        # given
        expected = TaskLogFactory.build(
            state=TaskLog.State.FAILURE, exception="", traceback=""
        )
        # when
        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = expected.timestamp
            result = TaskLog.objects.create_from_task(
                task_id=str(expected.task_id),
                task_name=expected.task_name,
                state=expected.state,
                priority=expected.priority,
                retries=expected.retries,
                received=expected.received,
                started=expected.started,
            )
        # then
        self._assert_equal_objs(expected, result)

    def _assert_equal_objs(self, expected, result):
        field_names = {
            field.name for field in TaskLog._meta.fields if field.name != "id"
        }
        for field_name in field_names:
            with self.subTest(field_name=field_name):
                self.assertEqual(
                    getattr(expected, field_name), getattr(result, field_name)
                )



# Generated by Django 4.0.7 on 2022-08-24 19:46

import uuid

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="TaskLog",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("app_name", models.CharField(db_index=True, max_length=255)),
                ("exception", models.TextField()),
                ("parent_id", models.UUIDField(default=None, null=True)),
                ("priority", models.IntegerField(default=None, null=True)),
                ("retries", models.IntegerField()),
                ("received", models.DateTimeField(default=None, null=True)),
                ("runtime", models.FloatField(db_index=True, default=None, null=True)),
                ("started", models.DateTimeField(default=None, null=True)),
                (
                    "state",
                    models.IntegerField(
                        choices=[(1, "success"), (2, "retry"), (3, "failure")],
                        db_index=True,
                    ),
                ),
                ("task_id", models.UUIDField(db_index=True, default=uuid.uuid4)),
                ("task_name", models.CharField(db_index=True, max_length=255)),
                ("timestamp", models.DateTimeField(db_index=True)),
                ("traceback", models.TextField()),
            ],
        ),
    ]



from django.core.cache import cache
from django.test import TestCase

from taskmonitor.core import cached_reports
from taskmonitor.models import TaskLog

from ..factories import TaskLogFactory


class TestCachedReports(TestCase):
    def setUp(self) -> None:
        cache.clear()

    def test_should_create_reports(self):
        # given
        log_1 = TaskLogFactory(state=TaskLog.State.SUCCESS)
        log_2 = TaskLogFactory(state=TaskLog.State.FAILURE)
        log_3 = TaskLogFactory(state=TaskLog.State.RETRY)
        # when
        result = cached_reports._calc_data()
        # then
        oldest = min(log_1.timestamp, log_2.timestamp, log_3.timestamp)
        self.assertEqual(result["oldest_date"], oldest)
        newest = max(log_1.timestamp, log_2.timestamp, log_3.timestamp)
        self.assertEqual(result["youngest_date"], newest)
        self.assertEqual(result["total_runs"], 3)



import uuid

from django.db import models

from .managers import TaskLogManager


class TaskReport(models.Model):
    """Dummy model to fake a 'Reports' entry on the admin index page."""

    id = models.BigIntegerField(primary_key=True)

    class Meta:
        managed = False
        verbose_name = "report"


class TaskLog(models.Model):
    """Log entry for an executed celery task."""

    class State(models.IntegerChoices):
        SUCCESS = 1, "success"
        RETRY = 2, "retry"
        FAILURE = 3, "failure"

    app_name = models.CharField(max_length=255, db_index=True)
    exception = models.TextField()
    parent_id = models.UUIDField(null=True, default=None)
    priority = models.IntegerField(null=True, default=None)
    retries = models.IntegerField()
    received = models.DateTimeField(null=True, default=None)
    runtime = models.FloatField(null=True, default=None, db_index=True)
    started = models.DateTimeField(null=True, default=None)
    state = models.IntegerField(choices=State.choices, db_index=True)
    task_id = models.UUIDField(default=uuid.uuid4, db_index=True)
    task_name = models.CharField(max_length=255, db_index=True)
    timestamp = models.DateTimeField(db_index=True)
    traceback = models.TextField()

    objects = TaskLogManager()

    def __str__(self):
        return f"{self.task_name}:{self.pk}"

    def save(self, *args, **kwargs) -> None:
        if self.started:
            self.runtime = (self.timestamp - self.started).total_seconds()
        super().save(*args, **kwargs)



from django.contrib.admin.sites import AdminSite

from app_utils.testing import NoSocketsTestCase

from taskmonitor import admin
from taskmonitor.models import TaskLog

from .factories import TaskLogFactory


class TestTaskLogAdmin(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.modeladmin = admin.TaskLogAdmin(model=TaskLog, admin_site=AdminSite())

    def test_should_format_runtime(self):
        # given
        log = TaskLogFactory()
        log.runtime = 0.2
        # when/then
        self.assertEqual(self.modeladmin._runtime(log), "0.2")

    def test_should_handle_no_runtime(self):
        # given
        log = TaskLogFactory()
        log.runtime = None
        # when/then
        self.assertIsNone(self.modeladmin._runtime(log))



import datetime as dt
from dataclasses import asdict, dataclass
from random import choice, choices, randint
from uuid import UUID

import factory
import factory.fuzzy
from factory.faker import faker

from django.utils import timezone

from taskmonitor.models import TaskLog

# generate fake apps and task names
faker = faker.Faker()
fake_tasks = {}
for app_name in {faker.first_name().lower() for _ in range(10)}:
    fake_tasks[app_name] = [
        app_name + ".tasks." + "_".join(faker.words(3)).lower()
        for _ in range(randint(3, 20))
    ]


class TaskLogFactory(factory.django.DjangoModelFactory):
    class Meta:
        model = TaskLog

    app_name = factory.fuzzy.FuzzyChoice(fake_tasks.keys())
    received = factory.fuzzy.FuzzyDateTime(timezone.now() - dt.timedelta(minutes=5))
    started = factory.LazyAttribute(
        lambda o: factory.fuzzy.FuzzyDateTime(start_dt=o.received).fuzz()
    )
    task_name = factory.LazyAttribute(lambda o: choice(fake_tasks[o.app_name]))

    @factory.lazy_attribute
    def task_id(self):
        return UUID(faker.uuid4())

    @factory.lazy_attribute
    def runtime(self):
        return (self.timestamp - self.started).total_seconds()

    @factory.lazy_attribute
    def priority(self):
        return choices(
            population=[1, 2, 3, 4, 5, 6, 7, 8, 9],
            weights=[5, 5, 10, 20, 100, 20, 10, 5, 5],
        )[0]

    @factory.lazy_attribute
    def retries(self):
        return choices(
            population=[0, 1, 2, 3],
            weights=[80, 10, 5, 5],
        )[0]

    @factory.lazy_attribute
    def state(self):
        return choices(
            population=[
                TaskLog.State.SUCCESS,
                TaskLog.State.RETRY,
                TaskLog.State.FAILURE,
            ],
            weights=[80, 15, 5],
        )[0]

    @factory.lazy_attribute
    def exception(self):
        if self.state == TaskLog.State.SUCCESS:
            return ""
        return faker.sentence()

    @factory.lazy_attribute
    def traceback(self):
        if self.state == TaskLog.State.SUCCESS:
            return ""
        return faker.paragraph()

    @factory.lazy_attribute
    def timestamp(self):
        max_duration = choices(
            population=[0.5, 1, 10, 30, 120],
            weights=[75, 10, 5, 5, 5],
        )[0]
        start_dt = self.started + dt.timedelta(seconds=0.1)
        return factory.fuzzy.FuzzyDateTime(
            start_dt=start_dt,
            end_dt=start_dt + dt.timedelta(seconds=max_duration),
        ).fuzz()


@dataclass
class ContextStub:
    id: str
    retries: int
    delivery_info: dict
    parent_id: str = None

    def asdict(self) -> dict:
        return asdict(self)

    @classmethod
    def create_from_obj(cls, obj: TaskLog):
        return cls(
            parent_id=obj.parent_id,
            retries=obj.retries,
            id=str(obj.task_id),
            delivery_info={
                "is_eager": False,
                "exchange": None,
                "routing_key": None,
                "priority": obj.priority,
            },
        )


@dataclass
class SenderStub:
    name: str
    request: ContextStub
    priority: int

    @classmethod
    def create_from_obj(cls, obj: TaskLog):
        request = ContextStub.create_from_obj(obj)
        return cls(name=obj.task_name, request=request, priority=5)



from django.urls import path

from . import views

app_name = "taskmonitor"

urlpatterns = [
    path(
        "admin_taskmonitor_download_csv",
        views.admin_taskmonitor_download_csv,
        name="admin_taskmonitor_download_csv",
    ),
    path(
        "admin_taskmonitor_reports",
        views.admin_taskmonitor_reports,
        name="admin_taskmonitor_reports",
    ),
    path(
        "admin_taskmonitor_reports_clear_cache",
        views.admin_taskmonitor_reports_clear_cache,
        name="admin_taskmonitor_reports_clear_cache",
    ),
]



"""Container for caching the reports data used in views."""

import datetime as dt
from typing import Optional

from django.core.cache import cache
from django.db import models
from django.db.models import Count, F, Max, Min, Sum, Value
from django.db.models.functions import Concat
from django.urls import reverse
from django.utils import timezone

from ..app_settings import (
    TASKMONITOR_HOUSEKEEPING_FREQUENCY,
    TASKMONITOR_REPORTS_MAX_AGE,
    TASKMONITOR_REPORTS_MAX_TOP,
)
from ..models import TaskLog

CACHE_KEY = "TASKMONITOR_REPORTS_DATA"


def data() -> dict:
    """Return the cached reports data."""
    context = cache.get_or_set(CACHE_KEY, _calc_data, timeout=_timeout())
    ttl = cache.ttl(CACHE_KEY)
    context["last_update_at"] = _last_update_at(ttl)
    context["next_update_at"] = _next_update_at(ttl)
    return context


def _timeout() -> int:
    """Timeout in seconds."""
    return TASKMONITOR_REPORTS_MAX_AGE * 60


def _last_update_at(ttl) -> Optional[dt.datetime]:
    """When the cache was last updated or None if there is no cache."""
    if not ttl:
        return None
    return timezone.now() - dt.timedelta(seconds=max(0, _timeout() - ttl))


def _next_update_at(ttl) -> Optional[dt.datetime]:
    """When the cache will be updated next (earliest) or None if no cache."""
    if not ttl:
        return None
    duration = TASKMONITOR_HOUSEKEEPING_FREQUENCY * 60 / 2 + ttl
    return timezone.now() + dt.timedelta(seconds=duration)


def refresh_cache() -> None:
    """Refresh the cache."""
    cache.set(CACHE_KEY, _calc_data(), timeout=_timeout())


def clear_cache() -> None:
    """Clear the cache."""
    cache.delete(CACHE_KEY)


def _calc_data() -> dict:
    """Calculate the report data."""
    oldest_date = TaskLog.objects.aggregate(oldest=Min("timestamp"))["oldest"]
    youngest_date = TaskLog.objects.aggregate(youngest=Max("timestamp"))["youngest"]
    total_runtime = TaskLog.objects.aggregate(total_runtime=Sum("runtime"))[
        "total_runtime"
    ]
    try:
        total_runtime_date = timezone.now() - dt.timedelta(seconds=total_runtime)
    except TypeError:
        total_runtime_date = None
    total_runs = TaskLog.objects.count()
    changelist_url = reverse("admin:taskmonitor_tasklog_changelist")
    task_totals_by_state = [
        {
            "name": state.label,
            "amount": (amount := TaskLog.objects.filter(state=state.value).count()),
            "percent": amount / total_runs * 100,
            "url": f"{changelist_url}?state__exact={state}",
        }
        for state in TaskLog.State
    ]
    task_runs_per_app = (
        TaskLog.objects.values(name=F("app_name"))
        .annotate(amount=Count("pk"))
        .annotate(
            percent=F("amount")
            / Value(total_runs, output_field=models.FloatField())
            * 100
        )
        .annotate(url=Concat(Value(f"{changelist_url}?app_name="), F("name")))
        .order_by("-amount")
    )
    tasks_top_runs = (
        TaskLog.objects.values(name=F("task_name"))
        .annotate(amount=Count("pk"))
        .annotate(
            percent=F("amount")
            / Value(total_runs, output_field=models.FloatField())
            * 100
        )
        .annotate(url=Concat(Value(f"{changelist_url}?task_name="), F("name")))
        .order_by("-amount")[:TASKMONITOR_REPORTS_MAX_TOP]
    )
    tasks_top_runtime = (
        TaskLog.objects.values(name=F("task_name"))
        .annotate(amount=Max("runtime"))
        .annotate(
            percent=F("amount")
            / Value(total_runtime, output_field=models.FloatField())
            * 100
        )
        .annotate(url=Concat(Value(f"{changelist_url}?o=5&task_name="), F("name")))
        .order_by("-amount")[:TASKMONITOR_REPORTS_MAX_TOP]
    )
    total_failed = TaskLog.objects.filter(state=TaskLog.State.FAILURE).count()
    tasks_top_failed = (
        TaskLog.objects.filter(state=TaskLog.State.FAILURE)
        .values(name=F("task_name"))
        .annotate(amount=Count("pk"))
        .annotate(
            percent=F("amount")
            / Value(total_failed, output_field=models.FloatField())
            * 100
        )
        .annotate(
            url=Concat(Value(f"{changelist_url}?state__exact=3&task_name="), F("name"))
        )
        .order_by("-amount")[:TASKMONITOR_REPORTS_MAX_TOP]
    )
    total_retried = TaskLog.objects.filter(state=TaskLog.State.RETRY).count()
    tasks_top_retried = (
        TaskLog.objects.filter(state=TaskLog.State.RETRY)
        .values(name=F("task_name"))
        .annotate(amount=Count("pk"))
        .annotate(
            percent=F("amount")
            / Value(total_retried, output_field=models.FloatField())
            * 100
        )
        .annotate(
            url=Concat(Value(f"{changelist_url}?state__exact=2&task_name="), F("name"))
        )
        .order_by("-amount")[:TASKMONITOR_REPORTS_MAX_TOP]
    )
    context = {
        "oldest_date": oldest_date,
        "youngest_date": youngest_date,
        "total_runs": total_runs,
        "total_runtime_date": total_runtime_date,
        "task_totals_by_state": task_totals_by_state,
        "task_runs_per_app": task_runs_per_app,
        "tasks_top_runs": tasks_top_runs,
        "tasks_top_runtime": tasks_top_runtime,
        "tasks_top_failed": tasks_top_failed,
        "tasks_top_retried": tasks_top_retried,
        "MAX_TOP": TASKMONITOR_REPORTS_MAX_TOP,
    }
    return context



import datetime as dt
from unittest.mock import patch

from django.test import TestCase
from django.utils import timezone

from taskmonitor.models import TaskLog
from taskmonitor.tasks import delete_stale_tasklogs

from .factories import TaskLogFactory

TASKS_PATH = "taskmonitor.tasks"


@patch(TASKS_PATH + ".TASKMONITOR_DATA_MAX_AGE", 3)
class TestTasks(TestCase):
    def test_should_delete_stale_entries_only(self):
        # given
        stale_entry = TaskLogFactory(
            timestamp=timezone.now() - dt.timedelta(hours=3, seconds=1)
        )
        current_entry = TaskLogFactory(timestamp=timezone.now())
        # when
        delete_stale_tasklogs()
        # then
        self.assertFalse(TaskLog.objects.filter(pk=stale_entry.pk).exists())
        self.assertTrue(TaskLog.objects.filter(pk=current_entry.pk).exists())



"""A global key/value store for task related data.

The django cache is used as data storage to enable persistence and thread safety.
"""

from typing import Any, Optional

from django.core.cache import cache

from ..app_settings import TASKMONITOR_DATA_MAX_AGE

CACHE_KEY = "TASKMONITOR_RECORDS"


def set(task_id: str, key: str, value: Any):
    """Set a key/value for a task."""
    timeout = TASKMONITOR_DATA_MAX_AGE * 3600
    cache.set(_build_key(task_id, key), value, timeout=timeout)


def get(task_id: str, key: str) -> Optional[Any]:
    """Returns the value for a task and key."""
    return cache.get(_build_key(task_id, key))


def delete(task_id: str, key: str) -> None:
    """Remove a key for a task."""
    cache.delete(_build_key(task_id, key))


def fetch(task_id: str, key: str) -> Optional[Any]:
    """Fetch the value and removes it's key for task."""
    value = get(task_id, key)
    delete(task_id, key)
    return value


def _build_key(task_id, key: str) -> str:
    return f"{CACHE_KEY}_{task_id}_{key}"



from django.test import TestCase

from taskmonitor.helpers import extract_app_name


class TestExtractAppName(TestCase):
    def test_can_extract_from_normal_task_name(self):
        # when
        result = extract_app_name("alpha.tasks.task_name")
        # then
        self.assertEqual(result, "alpha")

    def test_can_extract_from_long_task_name(self):
        # when
        result = extract_app_name("omega.alpha.tasks.task_name")
        # then
        self.assertEqual(result, "alpha")

    def test_can_extract_from_extra_long_task_name(self):
        # when
        result = extract_app_name("echo.omega.alpha.tasks.task_name")
        # then
        self.assertEqual(result, "alpha")

    def test_can_extract_from_custom_task_name(self):
        # when
        result = extract_app_name("alpha.task_name")
        # then
        self.assertEqual(result, "alpha")

    def test_should_return_empty_string_if_no_match_1(self):
        # when
        result = extract_app_name("dummy")
        # then
        self.assertEqual(result, "")

    def test_should_return_empty_string_if_no_match_2(self):
        # when
        result = extract_app_name("tasks.dummy")
        # then
        self.assertEqual(result, "")



import datetime as dt
import traceback as tb
from typing import List
from uuid import UUID

from django.db import models
from django.utils import timezone

from .helpers import extract_app_name


class TaskLogQuerySet(models.QuerySet):
    def csv_line_generator(self, fields: List[str]):
        """Return the tasklogs for a CSV file line by line.
        And return the field names as first line.
        """
        yield [field.name for field in fields]
        for obj in self.iterator():
            values = []
            for field in fields:
                if field.choices:
                    value = getattr(obj, f"get_{field.name}_display")()
                else:
                    value = getattr(obj, field.name)
                # if callable(value):
                #     try:
                #         value = value() or ""
                #     except Exception:
                #         value = "Error retrieving value"
                if value is None:
                    value = ""
                values.append(value)
            yield values


class TaskLogManagerBase(models.Manager):
    def create_from_task(
        self,
        *,
        task_id: str,
        task_name: str,
        state: int,
        retries: int,
        priority: int,
        received: dt.datetime = None,
        started: dt.datetime = None,
        parent_id: str = None,
        exception=None,
    ) -> models.Model:
        """Create new object from a celery task."""
        args = {
            "app_name": extract_app_name(task_name),
            "priority": priority,
            "parent_id": UUID(parent_id) if parent_id else None,
            "received": received,
            "retries": retries,
            "started": started,
            "state": state,
            "task_id": UUID(task_id),
            "task_name": task_name,
            "timestamp": timezone.now(),
        }
        if exception:
            args["exception"] = str(exception)
            if traceback := getattr(exception, "__traceback__"):
                args["traceback"] = "".join(
                    tb.format_exception(None, value=exception, tb=traceback)
                )
        return self.create(**args)


TaskLogManager = TaskLogManagerBase.from_queryset(TaskLogQuerySet)



import pytz

from django.test import RequestFactory, TestCase
from django.urls import reverse

from app_utils.testdata_factories import UserFactory

from taskmonitor import views
from taskmonitor.models import TaskLog

from .factories import TaskLogFactory


def _format_dt(my_dt) -> str:
    """Format datetime into a format with a special timezone offset,
    which strtime() does not support.
    """
    offset = pytz.utc.utcoffset(my_dt).total_seconds()
    sign = "+" if offset >= 0 else "-"
    hours = int(offset / 3600)
    minutes = int(offset / 60)
    offset_str = f"{sign}{hours:02}:{minutes:02}"
    return my_dt.strftime("%Y-%m-%d %H:%M:%S.%f") + offset_str


def _streaming_content_to_string(response):
    bytes = b"".join(response.streaming_content)
    return bytes.decode("utf-8")


class TestViews(TestCase):
    def setUp(self) -> None:
        self.request_factory = RequestFactory()

    def test_should_export_data_to_csv(self):
        # given
        user = UserFactory(is_staff=True)
        request = self.request_factory.get(
            reverse("taskmonitor:admin_taskmonitor_download_csv")
        )
        request.user = user
        self.maxDiff = None
        # when
        TaskLogFactory(state=TaskLog.State.SUCCESS)
        TaskLogFactory(state=TaskLog.State.RETRY)
        TaskLogFactory(state=TaskLog.State.FAILURE)
        for _ in range(50):
            TaskLogFactory()
        response = views.admin_taskmonitor_download_csv(request)
        # then
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.headers["Content-Type"], "text/csv")
        content = _streaming_content_to_string(response)
        lines = content.splitlines()
        lines.reverse()
        self.assertEqual(
            lines.pop(),
            "id;app_name;exception;parent_id;priority;retries;received;runtime;started;state;"
            "task_id;task_name;timestamp",
        )
        for entry in TaskLog.objects.order_by("pk"):
            values = lines.pop().split(";")
            self.assertEqual(len(values), 13)
            self.assertEqual(values[0], str(entry.pk))
            self.assertEqual(values[1], entry.app_name)
            self.assertEqual(values[2], "" if not entry.exception else entry.exception)
            self.assertEqual(values[3], "" if not entry.parent_id else entry.parent_id)
            self.assertEqual(values[4], str(entry.priority))
            self.assertEqual(values[5], str(entry.retries))
            self.assertEqual(values[6], _format_dt(entry.received))
            self.assertEqual(values[7], str(entry.runtime))
            self.assertEqual(values[8], _format_dt(entry.started))
            self.assertEqual(values[9], entry.get_state_display())
            self.assertEqual(values[10], str(entry.task_id))
            self.assertEqual(values[11], entry.task_name)
            self.assertEqual(values[12], _format_dt(entry.timestamp))



"""Create tasklogs from executed celery tasks."""

from django.core.cache import cache
from django.utils import timezone

from ..app_settings import TASKMONITOR_HOUSEKEEPING_FREQUENCY
from ..models import TaskLog
from ..tasks import DEFAULT_TASK_PRIORITY, run_housekeeping
from . import task_records

TASK_RECEIVED = "received"
TASK_STARTED = "started"

CACHE_KEY = "TASKMONITOR_LAST_HOUSEKEEPING"


def run_housekeeping_if_stale():
    """Spawn a task to run house keeping if last run was too long ago."""
    was_expired = cache.add(
        key=CACHE_KEY,
        value="no-value",
        timeout=TASKMONITOR_HOUSEKEEPING_FREQUENCY * 60,
    )
    if was_expired:
        run_housekeeping.apply_async(priority=DEFAULT_TASK_PRIORITY)


def task_received_handler_2(request):
    """Handle task received signal."""
    if request:
        task_records.set(request.id, TASK_RECEIVED, timezone.now())


def task_prerun_handler_2(task_id):
    """Handle task prerun signal."""
    if task_id:
        task_records.set(task_id, TASK_STARTED, timezone.now())


def task_success_handler_2(sender):
    """Handle task success signal."""
    if sender and sender.request:
        task_id = sender.request.id
        TaskLog.objects.create_from_task(
            task_id=task_id,
            task_name=sender.name,
            state=TaskLog.State.SUCCESS,
            retries=sender.request.retries,
            priority=sender.request.delivery_info["priority"],
            parent_id=sender.request.parent_id,
            received=task_records.fetch(task_id, TASK_RECEIVED),
            started=task_records.fetch(task_id, TASK_STARTED),
        )
    run_housekeeping_if_stale()


def task_retry_handler_2(sender, request, reason):
    """Handle task retry signal."""
    if sender and request:
        task_id = request.id
        TaskLog.objects.create_from_task(
            task_id=task_id,
            task_name=sender.name,
            state=TaskLog.State.RETRY,
            retries=request.retries,
            priority=request.delivery_info["priority"],
            parent_id=request.parent_id,
            received=task_records.fetch(task_id, TASK_RECEIVED),
            started=task_records.fetch(task_id, TASK_STARTED),
            exception=reason,
        )
    run_housekeeping_if_stale()


def task_failure_handler_2(sender, task_id, exception):
    """Handle task failure signal."""
    if sender and task_id:
        TaskLog.objects.create_from_task(
            task_id=task_id,
            task_name=sender.name,
            state=TaskLog.State.FAILURE,
            retries=sender.request.retries,
            priority=sender.request.delivery_info["priority"],
            parent_id=sender.request.parent_id,
            received=task_records.fetch(task_id, TASK_RECEIVED),
            started=task_records.fetch(task_id, TASK_STARTED),
            exception=exception,
        )
    run_housekeeping_if_stale()


def task_internal_error_handler_2(sender, task_id, request, exception):
    """Handle task internal error signal."""
    if task_id and request:
        TaskLog.objects.create_from_task(
            task_id=task_id,
            task_name=sender.name if sender else "?",
            state=TaskLog.State.FAILURE,
            retries=request["retries"],
            priority=request["delivery_info"].get("priority"),
            parent_id=request.get("parent_id"),
            received=task_records.fetch(task_id, TASK_RECEIVED),
            started=task_records.fetch(task_id, TASK_STARTED),
            exception=exception,
        )
    run_housekeeping_if_stale()


# def request_asdict(request) -> dict:
#     """Convert a request object into a dict."""
#     return {
#         "delivery_info": request.delivery_info,
#         "retries": request.retries,
#         "parent_id": request.parent_id,
#     }



import os

from setuptools import find_packages, setup

from taskmonitor import __version__

with open(os.path.join(os.path.dirname(__file__), "README.md")) as readme:
    README = readme.read()

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name="aa-taskmonitor",
    version=__version__,
    packages=find_packages(),
    include_package_data=True,
    license="MIT",
    description="Alliance Auth plugin for monitoring celery tasks",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/ErikKalkoken/aa-taskmonitor",
    author="Erik Kalkoken",
    author_email="kalkoken87@gmail.com",
    classifiers=[
        "Environment :: Web Environment",
        "Framework :: Django",
        "Framework :: Django :: 3.2",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
    ],
    python_requires="~=3.8",
    install_requires=[
        "allianceauth>=2.9",
        "allianceauth-app-utils>=1.14.2",
    ],
)



from typing import Optional

from django.contrib import admin
from django.shortcuts import redirect
from django.utils import html

from .models import TaskLog, TaskReport


@admin.register(TaskReport)
class TaskReportAdmin(admin.ModelAdmin):
    def has_add_permission(self, *args, **kwargs):
        return False

    def has_change_permission(self, *args, **kwargs):
        return False

    def has_delete_permission(self, *args, **kwargs):
        return False

    def changelist_view(self, request, extra_context=None):
        return redirect("taskmonitor:admin_taskmonitor_reports")


@admin.register(TaskLog)
class TaskLogAdmin(admin.ModelAdmin):
    class Media:
        css = {"all": ("taskmonitor/admin.css",)}

    list_display = (
        "timestamp",
        "task_name",
        "priority",
        "_state",
        "_runtime",
        "_exception",
    )
    list_filter = ("state", "timestamp", "app_name", "task_name")
    search_fields = ("task_name", "app_name", "task_id")
    actions = ["delete_selected_2"]

    def has_add_permission(self, *args, **kwargs) -> bool:
        return False

    def has_change_permission(self, *args, **kwargs):
        return False

    def get_actions(self, request):
        actions = super().get_actions(request)
        if "delete_selected" in actions:
            del actions["delete_selected"]
        return actions

    @admin.display(ordering="runtime")
    def _runtime(self, obj) -> Optional[str]:
        return f"{obj.runtime:.1f}" if obj.runtime else None

    @admin.display(ordering="state")
    def _state(self, obj) -> str:
        css_class_map = {
            TaskLog.State.RETRY: "state-retry",
            TaskLog.State.FAILURE: "state-failure",
        }
        css_class = css_class_map.get(obj.state, "")
        return html.format_html(
            '<span class="{}">{}</span>', css_class, obj.get_state_display()
        )

    @admin.display(ordering="exception")
    def _exception(self, obj) -> str:
        return obj.exception
        # if obj.exception:
        #     return html.format_html(
        #         '<span class="truncate" title="{}">{}</span>',
        #         obj.exception,
        #         obj.exception,
        #     )
        # return ""

    @admin.action(description="Delete selected entries (NO CONFIRMATION!")
    def delete_selected_2(self, request, queryset):
        entries_count = queryset.count()
        queryset._raw_delete(queryset.db)
        self.message_user(request, f"Deleted {entries_count} entries.")



class Echo:
    """An object that implements just the write method of the file-like
    interface.
    """

    def write(self, value):
        """Write the value by returning it, instead of storing in a buffer."""
        return value


def extract_app_name(task_name: str) -> str:
    """Extract the app name from a typical task name."""
    parts = task_name.split(".")
    try:
        idx = parts.index("tasks")
    except ValueError:
        if len(parts) == 2:
            return parts[0]
        else:
            return ""
    return parts[idx - 1] if idx > 0 else ""



import datetime as dt

from celery import shared_task

from django.utils import timezone

from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import TASKMONITOR_DATA_MAX_AGE
from .core import cached_reports
from .models import TaskLog

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

DEFAULT_TASK_PRIORITY = 7


@shared_task(base=QueueOnce)
def run_housekeeping():
    """Run all housekeeping tasks."""
    delete_stale_tasklogs.apply_async(priority=DEFAULT_TASK_PRIORITY)
    refresh_reports_cache.apply_async(priority=DEFAULT_TASK_PRIORITY)


@shared_task
def delete_stale_tasklogs():
    """Delete all stale tasklogs from the database."""
    old_entries = TaskLog.objects.filter(
        timestamp__lte=timezone.now() - dt.timedelta(hours=TASKMONITOR_DATA_MAX_AGE)
    )
    old_entries_count = old_entries.count()
    old_entries._raw_delete(old_entries.db)
    logger.info(f"Deleted {old_entries_count:,} stale tasklogs.")


@shared_task
def refresh_reports_cache():
    cached_reports.refresh_cache()
    logger.info("Refreshed the reports cache.")
